"""
The module that provides generating methods for all parts
of the pdelab driver.

Currently, these are hardcoded as strings. It would be possible
to switch these to cgen expression. OTOH, there is not much to be
gained there.
"""
from dune.perftool.generation import (generator_factory,
                                      include_file,
                                      symbol,
                                      preamble,
                                      )


# Those symbol generators that depend on the meta data we hacked into
# the ufl.FiniteElement should use fem_metadata_dependent_symbol instead of symbol.
def get_constraints_metadata(element):
    """
    Normally, we suppress the constraints information on a finite element.
    However, we might want to use it explicitly in those case where it matters
    This function does the job!
    """
    dc = getattr(element, 'dirichlet_constraints', None)
    de = getattr(element, 'dirichlet_expression', None)
    return (element, (dc, de))


fem_metadata_dependent_symbol = generator_factory(item_tags=("symbol",),
                                                  cache_key_generator=get_constraints_metadata,
                                                  )


# Have a global variable with the entire form data. This allows functions that depend
# deterministically on the entire data set to directly access it instead of passing it
# through the entire generator chain.
_form = None


# Have a function access this global data structure
def set_form(f):
    global _form
    _form = f


def isLagrange(fem):
    return fem._short_name is 'CG'


def isSimplical(fem):
    return any(fem._cell._cellname in x for x in ["triangle", "tetrahedron"])


def isQuadrilateral(fem):
    return any(fem._cell._cellname in x for x in ["vertex", "interval", "quadrilateral", "hexalateral"])


def isPk(fem):
    return isLagrange(fem) and isSimplical(fem)


def isQk(fem):
    return isLagrange(fem) and isQuadrilateral(fem)


def FEM_name_mangling(fem):
    from ufl import MixedElement, VectorElement, FiniteElement
    if isinstance(fem, MixedElement):
        name = ""
        for elem in fem._sub_elements:
            if name is not "":
                name = name + "_"
            name = name + FEM_name_mangling(elem)
        return name
    if isinstance(fem, VectorElement):
        return FEM_name_mangling(fem._sub_elements[0]) + "_" + str(fem._cell._geometric_dimension)
    if isinstance(fem, FiniteElement):
        if isPk(fem):
            return "P" + str(fem._degree)
        if isQk(fem):
            return "Q" + str(fem._degree)
    raise NotImplementedError("FEM NAME MANGLING")


@symbol
def name_inifile():
    # TODO pass some other option here.
    return "argv[1]"


@preamble
def parse_initree(varname):
    include_file("dune/common/parametertree.hh", filetag="driver")
    include_file("dune/common/parametertreeparser.hh", filetag="driver")
    filename = name_inifile()
    return ["Dune::ParameterTree initree;", "Dune::ParameterTreeParser::readINITree({}, {});".format(filename, varname)]


@symbol
def name_initree():
    parse_initree("initree")
    # TODO we can get some other ini file here.
    return "initree"


@preamble
def define_dimension(name):
    return "static const int {} = {};".format(name, _form.cell().geometric_dimension())


@symbol
def name_dimension():
    define_dimension("dim")
    return "dim"


@preamble
def typedef_grid(name):
    dim = name_dimension()
    if any(_form.cell().cellname() in x for x in ["vertex", "interval", "quadrilateral", "hexalateral"]):
        gridt = "Dune::YaspGrid<{}>".format(dim)
        include_file("dune/grid/yaspgrid.hh", filetag="driver")
    else:
        if any(_form.cell().cellname() in x for x in ["triangle", "tetrahedron"]):
            gridt = "Dune::UGGrid<{}>".format(dim)
            include_file("dune/grid/uggrid.hh", filetag="driver")
        else:
            raise ValueError("Cant match your geometry with a DUNE grid. Please report bug.")
    return "typedef {} {};".format(gridt, name)


@symbol
def type_grid():
    typedef_grid("Grid")
    return "Grid"


@preamble
def define_grid(name):
    include_file("dune/testtools/gridconstruction.hh", filetag="driver")
    ini = name_initree()
    _type = type_grid()
    return ["IniGridFactory<{}> factory({});".format(_type, ini),
            "std::shared_ptr<{}> grid = factory.getGrid();".format(_type)]


@symbol
def name_grid():
    define_grid("grid")
    return "grid"


@preamble
def typedef_leafview(name):
    grid = type_grid()
    return "typedef {}::LeafGridView {};".format(grid, name)


@symbol
def type_leafview():
    typedef_leafview("GV")
    return "GV"


@preamble
def define_leafview(name):
    _type = type_leafview()
    grid = name_grid()
    return "{} {} = {}->leafGridView();".format(_type, name, grid)


@symbol
def name_leafview():
    define_leafview("gv")
    return "gv"


@preamble
def typedef_vtkwriter(name):
    include_file("dune/grid/io/file/vtk/subsamplingvtkwriter.hh", filetag="driver")
    gv = type_leafview()
    return "typedef Dune::SubsamplingVTKWriter<{}> {};".format(gv, name)


@symbol
def type_vtkwriter():
    typedef_vtkwriter("VTKWriter")
    return "VTKWriter"


@preamble
def define_subsamplinglevel(name):
    ini = name_initree()
    return "int {} = {}.get<int>(\"vtk.subsamplinglevel\", 0);".format(name, ini)


@symbol
def name_subsamplinglevel():
    define_subsamplinglevel("sublevel")
    return "sublevel"


@preamble
def define_vtkwriter(name):
    _type = type_vtkwriter()
    gv = name_leafview()
    subsamp = name_subsamplinglevel()
    return "{} {}({}, {});".format(_type, name, gv, subsamp)


@symbol
def name_vtkwriter():
    define_vtkwriter("vtkwriter")
    return "vtkwriter"


@preamble
def typedef_domainfield(name):
    gridt = type_grid()
    return "typedef {}::ctype {};".format(gridt, name)


@symbol
def type_domainfield():
    typedef_domainfield("DF")
    return "DF"


@preamble
def typedef_range(name):
    return "typedef double {};".format(name)


@symbol
def type_range():
    typedef_range("R")
    return "R"


@preamble
def typedef_fem(expr, name):
    gv = type_leafview()
    df = type_domainfield()
    r = type_range()
    if isPk(expr):
        include_file("dune/pdelab/finiteelementmap/pkfem.hh", filetag="driver")
        return "typedef Dune::PDELab::PkLocalFiniteElementMap<{}, {}, {}, {}> {};".format(gv, df, r, expr._degree, name)
    if isQk(generator._kwargs['expr']):
        include_file("dune/pdelab/finiteelementmap/qkfem.hh", filetag="driver")
        return "typedef Dune::PDELab::QkLocalFiniteElementMap<{}, {}, {}, {}> {};".format(gv, df, r, expr._degree, name)
    raise NotImplementedError("FEM not implemented in dune-perftool")


@symbol
def type_fem(expr):
    name = "{}_FEM".format(FEM_name_mangling(expr).upper())
    typedef_fem(expr, name)
    return name


@preamble
def define_fem(expr, name):
    femtype = type_fem(expr)
    gv = name_leafview()
    return "{} {}({});".format(femtype, name, gv)


@symbol
def name_fem(expr):
    name = "{}_fem".format(FEM_name_mangling(expr).lower())
    define_fem(expr, name)
    return name


@preamble
def typedef_vectorbackend(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::istl::VectorBackend<Dune::PDELab::istl::Blocking::none, 1> {};".format(name)


@symbol
def type_vectorbackend():
    typedef_vectorbackend("VectorBackend")
    return "VectorBackend"


@symbol
def type_orderingtag():
    return "Dune::PDELab::LexicographicOrderingTag"


@preamble
def typedef_dirichlet_constraintsassembler(name):
    include_file("dune/pdelab/constraints/conforming.hh", filetag="driver")
    return "typedef Dune::PDELab::ConformingDirichletConstraints {};".format(name)


@preamble
def typedef_no_constraintsassembler(name):
    return "typedef Dune::PDELab::NoConstraints {};".format(name)


@symbol
def type_constraintsassembler(dirichlet):
    if dirichlet:
        typedef_dirichlet_constraintsassembler("DirichletConstraintsAssember")
        return "DirichletConstraintsAssember"
    else:
        typedef_no_constraintsassembler("NoConstraintsAssembler")
        return "NoConstraintsAssembler"


@preamble
def typedef_constraintscontainer(element, name):
    gfs = type_gfs(element)
    r = type_range()
    return "typedef {}::ConstraintsContainer<{}>::Type {};".format(gfs, r, name)


@symbol
def type_constraintscontainer(element):
    name = "{}_cc".format(FEM_name_mangling(element)).upper()
    typedef_constraintscontainer(element, name)
    return name


@preamble
def define_constraintscontainer(expr, name):
    cctype = type_constraintscontainer(expr)
    return ["{} {};".format(cctype, name), "{}.clear();".format(name)]


@symbol
def name_constraintscontainer(element, dirichlet):
    name = "{}{}_cc".format(FEM_name_mangling(element), "_dirichlet" if dirichlet else "").lower()
    define_constraintscontainer(element, name)
    return name


@preamble
def define_intersection_lambda(expression, name):
    return "auto {} = [&](const auto& is, const auto& x){{ {} }};".format(name, expression.c_expr)


@symbol
def name_bctype_lambda(dirichlet):
    define_intersection_lambda(dirichlet, "glambda")
    # TODO get access to the name data dict here!
    return "glambda"


@preamble
def define_bctype_function(dirichlet, name):
    gv = name_leafview()
    bctype_lambda = name_bctype_lambda(dirichlet)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    return "auto {} = Dune::PDELab::makeBoundaryConditionFromCallable({}, {});".format(name,
                                                                                       gv,
                                                                                       bctype_lambda,
                                                                                       )


@symbol
def name_bctype_function(dirichlet):
    define_bctype_function(dirichlet, "g")
    # TODO get access to the name data dict here!
    return "g"


@preamble
def assemble_constraints(name, element, dirichlet):
    assert dirichlet
    bctype_function = name_bctype_function(dirichlet)
    gfs = name_gfs(element)
    return "Dune::PDELab::constraints({}, {}, {});".format(bctype_function,
                                                           gfs,
                                                           name,
                                                           )


@fem_metadata_dependent_symbol
def name_assembled_constraints(expr):
    element, (dirichlet, _) = get_constraints_metadata(expr)
    name = name_constraintscontainer(element, dirichlet)
    if dirichlet:
        assemble_constraints(name, element, dirichlet)
    return name


@preamble
def typedef_gfs(element, dirichlet, name):
    vb = type_vectorbackend()
    from ufl import FiniteElement, MixedElement, VectorElement, EnrichedElement, RestrictedElement
    if isinstance(element, FiniteElement):
        gv = type_leafview()
        fem = type_fem(element)
        cass = type_constraintsassembler(dirichlet)
        return "typedef Dune::PDELab::GridFunctionSpace<{}, {}, {}, {}> {};".format(gv, fem, cass, vb, name)
    if isinstance(expr, MixedElement):
        ot = type_orderingtag()
        args = ", ".join(type_gfs(e) for e in expr._sub_elements)
        return "typedef Dune::PDELab::CompositeGridFunctionSpace<{}, {}, {}> {}".format(vb, ot, args, name)
    if isinstance(expr, VectorElement):
        include_file("dune/pdelab/gridfunctionspace/vectorgridfunctionspace.hh", filetag="driver")
        gv = type_leafview()
        fem = type_fem(expr._sub_elements[0])
        dim = name_dimension()
        cass = type_constraintsassembler(dirichlet)
        return "typedef Dune::PDELab::GridFunctionSpace<{}, {}, {}, {}, {}, {}> {};".format(gv, fem, dim, vb, vb, cass, name)
    if isinstance(expr, EnrichedElement):
        raise NotImplementedError("Dune does not support enriched elements!")
    if isinstance(expr, RestrictedElement):
        raise NotImplementedError("Dune does not support restricted elements!")


@fem_metadata_dependent_symbol
def type_gfs(expr):
    element, dirichlet = get_constraints_metadata(expr)
    name = "{}{}_gfs".format(FEM_name_mangling(element), "_dirichlet" if dirichlet else "").upper()
    typedef_gfs(element, dirichlet, name)
    return name


@preamble
def define_gfs(element, dirichlet, name):
    gfstype = type_gfs(element)
    from ufl import FiniteElement, MixedElement, VectorElement, EnrichedElement, RestrictedElement
    if isinstance(element, FiniteElement):
        gv = name_leafview()
        fem = name_fem(element)
        return "{} {}({}, {});".format(gfstype, name, gv, fem)
    if isinstance(expr, MixedElement):
        args = ", ".join(name_gfs(childgfs) for childgfs in element._sub_elements)
        return "{} {}({});".format(gfstype, name, args)
    if isinstance(expr, VectorElement):
        gv = name_leafview()
        fem = name_fem(element._sub_elements[0])
        return "{} {}({}, {});".format(gfstype, name, gv, fem)
    raise NotImplementedError("Dune does not support this type of finite element!")


@fem_metadata_dependent_symbol
def name_gfs(expr):
    element, dirichlet = get_constraints_metadata(expr)
    name = "{}{}_gfs".format(FEM_name_mangling(element), "_dirichlet" if dirichlet else "").lower()
    define_gfs(element, dirichlet, name)
    return name


@preamble
def define_dofestimate(name):
    # Provide a worstcase estimate for the number of entries per row based on the given gridfunction space and cell geometry
    if isQuadrilateral(_form.coefficients()[0].element()):
        geo_factor = "4"
    else:
        geo_factor = "6"
    gfs = name_gfs(_form.coefficients()[0].element())
    ini = name_initree()
    return ["int generic_dof_estimate =  {} * {}.maxLocalSize();".format(geo_factor, gfs),
            "int {} = {}.get<int>(\"istl.number_of_nnz\", generic_dof_estimate);".format(name, ini)]


@symbol
def name_dofestimate():
    define_dofestimate("dofestimate")
    return "dofestimate"


@preamble
def typedef_matrixbackend(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::istl::BCRSMatrixBackend<> {};".format(name)


@symbol
def type_matrixbackend():
    typedef_matrixbackend("MatrixBackend")
    return "MatrixBackend"


@preamble
def define_matrixbackend(name):
    mbtype = type_matrixbackend()
    dof = name_dofestimate()
    return "{} {}({});".format(mbtype, name, dof)


@symbol
def name_matrixbackend():
    define_matrixbackend("mb")
    return "mb"


@symbol
def type_parameters():
    return "LocalOperatorParams"


@preamble
def define_parameters(name):
    partype = type_parameters()
    return "{} {};".format(partype, name)


@symbol
def name_parameters():
    define_parameters("params")
    return "params"


@preamble
def typedef_localoperator(name):
    # No Parameter class here, yet
    # params = type_parameters()
    # return "typedef LocalOperator<{}> {};".format(params, name)
    ugfs = type_gfs(_form.coefficients()[0].element())
    vgfs = type_gfs(_form.arguments()[0].element())
    from dune.perftool.options import get_option
    include_file(get_option('operator_file'), filetag="driver")
    return "using {} = LocalOperator<{}, {}>;".format(name, ugfs, vgfs)


@symbol
def type_localoperator():
    typedef_localoperator("LOP")
    return "LOP"


@preamble
def define_localoperator(name):
    loptype = type_localoperator()
    ini = name_initree()
    params = name_parameters()
    return "{} {}({}, {});".format(loptype, name, ini, params)


@symbol
def name_localoperator():
    define_localoperator("lop")
    return "lop"


@preamble
def typedef_gridoperator(name):
    ugfs = type_gfs(_form.coefficients()[0].element())
    vgfs = type_gfs(_form.arguments()[0].element())
    lop = type_localoperator()
    ucc = type_constraintscontainer(_form.coefficients()[0].element())
    vcc = type_constraintscontainer(_form.arguments()[0].element())
    mb = type_matrixbackend()
    df = type_domainfield()
    r = type_range()
    include_file("dune/pdelab/gridoperator/gridoperator.hh", filetag="driver")
    return "typedef Dune::PDELab::GridOperator<{}, {}, {}, {}, {}, {}, {}, {}, {}> {};".format(ugfs, vgfs, lop, mb, df, r, r, ucc, vcc, name)


@symbol
def type_gridoperator():
    typedef_gridoperator("GO")
    return "GO"


@preamble
def define_gridoperator(name):
    gotype = type_gridoperator()
    ugfs = name_gfs(_form.coefficients()[0].element())
    ucc = name_assembled_constraints(_form.coefficients()[0].element())
    vgfs = name_gfs(_form.arguments()[0].element())
    vcc = name_assembled_constraints(_form.arguments()[0].element())
    lop = name_localoperator()
    mb = name_matrixbackend()
    return ["{} {}({}, {}, {}, {}, {}, {});".format(gotype, name, ugfs, ucc, vgfs, vcc, lop, mb),
            "std::cout << \"gfs with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(ugfs),
            "std::cout << \"cc with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(ucc)]


@symbol
def name_gridoperator():
    define_gridoperator("go")
    return "go"


@preamble
def typedef_vector(name):
    gotype = type_gridoperator()
    return "typedef {}::Traits::Domain {};".format(gotype, name)


@symbol
def type_vector():
    typedef_vector("V")
    return "V"


@preamble
def define_vector(name):
    vtype = type_vector()
    gfs = name_gfs(_form.coefficients()[0].element())
    return ["{} {}({});".format(vtype, name, gfs), "{} = 0.0;".format(name)]


@preamble
def define_boundary_lambda(boundary, name):
    return "auto {} = [&](const auto& x){{ {} }};".format(name, boundary.c_expr)


@symbol
def name_boundary_lambda(boundary, name):
    define_boundary_lambda(boundary, name + "lambda")
    return name + "lambda"


@preamble
def define_boundary_function(boundary, name):
    gv = name_leafview()
    lambdaname = name_boundary_lambda(boundary, name)
    return "auto {} = Dune::PDELab::makeGridFunctionFromCallable({}, {});".format(name,
                                                                                  gv,
                                                                                  lambdaname,
                                                                                  )


@symbol
def name_boundary_function(boundary):
    define_boundary_function(boundary, "b")
    # TODO add namespace data here
    return "b"


@preamble
def interpolate_vector(name):
    define_vector(name)
    element, (_, boundary) = get_constraints_metadata(_form.coefficients()[0].element())
    b = name_boundary_function(boundary)
    gfs = name_gfs(element)
    return "Dune::PDELab::interpolate({}, {}, {});".format(b,
                                                           gfs,
                                                           name,
                                                           )


@symbol
def name_vector():
    interpolate_vector("x")
    return "x"


@preamble
def typedef_linearsolver(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "typedef Dune::PDELab::ISTLBackend_SEQ_UMFPack {};".format(name)


@symbol
def type_linearsolver():
    typedef_linearsolver("LinearSolver")
    return "LinearSolver"


@preamble
def define_linearsolver(name):
    lstype = type_linearsolver()
    return "{} {}(false);".format(lstype, name)


@symbol
def name_linearsolver():
    define_linearsolver("ls")
    return "ls"


@preamble
def define_reduction(name):
    ini = name_initree()
    return "double {} = {}.get<double>(\"reduction\", 1e-12);".format(name, ini)


@symbol
def name_reduction():
    define_reduction("reduction")
    return "reduction"


@preamble
def typedef_stationarylinearproblemsolver(name):
    include_file("dune/pdelab/stationary/linearproblem.hh", filetag="driver")
    gotype = type_gridoperator()
    lstype = type_linearsolver()
    xtype = type_vector()
    return "typedef Dune::PDELab::StationaryLinearProblemSolver<{}, {}, {}> {};".format(gotype, lstype, xtype, name)


@symbol
def type_stationarylinearproblemsolver():
    typedef_stationarylinearproblemsolver("SLP")
    return "SLP"


@preamble
def define_stationarylinearproblemsolver(name):
    slptype = type_stationarylinearproblemsolver()
    go = name_gridoperator()
    ls = name_linearsolver()
    x = name_vector()
    red = name_reduction()
    return "{} {}({}, {}, {}, {});".format(slptype, name, go, ls, x, red)


@symbol
def name_stationarylinearproblemsolver():
    define_stationarylinearproblemsolver("slp")
    return "slp"


@preamble
def dune_solve():
    from ufl.algorithms.predicates import is_multilinear
    # This is crap as it does check for linearity of the rank 1 form,
    # which is by definition True.
    if is_multilinear(_form):
        slp = name_stationarylinearproblemsolver()
        return "{}.apply();".format(slp)
    else:
        raise NotImplementedError


@preamble
def define_vtkfile(name):
    ini = name_initree()
    include_file("string", filetag="driver")
    return "std::string {} = {}.get<std::string>(\"wrapper.vtkcompare.name\", \"output\");".format(name, ini)


@symbol
def name_vtkfile():
    define_vtkfile("vtkfile")
    return "vtkfile"


@preamble
def print_matrix():
    t_go = type_gridoperator()
    n_go = name_gridoperator()
    v = name_vector()
    return ["typedef typename {}::Traits::Jacobian M;".format(t_go),
            "M m({});".format(n_go),
            "{}.jacobian({},m);".format(n_go, v),
            "using Dune::PDELab::Backend::native;",
            "Dune::printmatrix(std::cout, native(m),\"global stiffness matrix\",\"row\",9,1);"]


@preamble
def define_gfs_name(element):
    gfs = name_gfs(element)
    #TODO make something sensible here
    return "{}.name(\"bla\");".format(gfs)


@preamble
def vtkoutput():
    element = _form.coefficients()[0].element()
    define_gfs_name(element)
    include_file("dune/pdelab/gridfunctionspace/vtk.hh", filetag="driver")
    vtkwriter = name_vtkwriter()
    gfs = name_gfs(element)
    vec = name_vector()
    vtkfile = name_vtkfile()
    dune_solve()
    #print_matrix()
    return ["Dune::PDELab::addSolutionToVTKWriter({}, {}, {});".format(vtkwriter, gfs, vec),
            "{}.write({}, Dune::VTK::ascii);".format(vtkwriter, vtkfile)]
